#include <algorithm>
#include <cmath>
#include <cstddef>
#include <memory>
#include <arbor/mechanism_abi.h>
#include <arbor/math.hpp>

namespace arb {
namespace default_catalogue {
namespace kernel_kamt {

using ::arb::math::exprelr;
using ::arb::math::safeinv;
using ::std::abs;
using ::std::cos;
using ::std::exp;
using ::std::max;
using ::std::min;
using ::std::pow;
using ::std::sin;
using ::std::sqrt;
using ::std::tanh;

static constexpr unsigned simd_width_ = 1;
static constexpr unsigned min_align_ = std::max(alignof(arb_value_type), alignof(arb_index_type));
using ::std::log;

#define PPACK_IFACE_BLOCK \
[[maybe_unused]] auto _pp_var_width                                                 = pp->width;\
[[maybe_unused]] auto _pp_var_n_detectors                                           = pp->n_detectors;\
[[maybe_unused]] auto _pp_var_dt                                                    = pp->dt;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_vec_ci                       = pp->vec_ci;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_v                        = pp->vec_v;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_i                        = pp->vec_i;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_g                        = pp->vec_g;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_temperature_degC             = pp->temperature_degC;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_diam_um                      = pp->diam_um;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_area_um2                     = pp->area_um2;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_time_since_spike             = pp->time_since_spike;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_node_index                   = pp->node_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_peer_index                   = pp->peer_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_multiplicity                 = pp->multiplicity;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_weight                       = pp->weight;\
[[maybe_unused]] auto& _pp_var_events                                               = pp->events;\
[[maybe_unused]] auto _pp_var_mechanism_id                                          = pp->mechanism_id;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_contiguous               = pp->index_constraints.n_contiguous;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_constant                 = pp->index_constraints.n_constant;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_independent              = pp->index_constraints.n_independent;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_none                     = pp->index_constraints.n_none;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_contiguous  = pp->index_constraints.contiguous;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_constant    = pp->index_constraints.constant;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_independent = pp->index_constraints.independent;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_none        = pp->index_constraints.none;\
[[maybe_unused]] auto _pp_var_a0m = pp->globals[0];\
[[maybe_unused]] auto _pp_var_vhalfm = pp->globals[1];\
[[maybe_unused]] auto _pp_var_zetam = pp->globals[2];\
[[maybe_unused]] auto _pp_var_gmm = pp->globals[3];\
[[maybe_unused]] auto _pp_var_a0h = pp->globals[4];\
[[maybe_unused]] auto _pp_var_vhalfh = pp->globals[5];\
[[maybe_unused]] auto _pp_var_zetah = pp->globals[6];\
[[maybe_unused]] auto _pp_var_gmh = pp->globals[7];\
[[maybe_unused]] auto _pp_var_sha = pp->globals[8];\
[[maybe_unused]] auto _pp_var_shi = pp->globals[9];\
[[maybe_unused]] auto const * const * _pp_var_random_numbers = pp->random_numbers;\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_m = pp->state_vars[0];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_h = pp->state_vars[1];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_v = pp->state_vars[2];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_gbar = pp->parameters[0];\
[[maybe_unused]] auto& _pp_var_ion_k = pp->ion_states[0];\
[[maybe_unused]] auto* __restrict__ _pp_var_ion_k_index = pp->ion_states[0].index;\
//End of IFACEBLOCK


// interface methods
static void init(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        _pp_var_m[i_] =  1.0/( 1.0+exp( -(v-_pp_var_sha- 7.5999999999999996)* 0.071428571428571425));
        _pp_var_h[i_] =  1.0/( 1.0+exp((v-_pp_var_shi+ 47.399999999999999)* 0.16666666666666666));
    }
    if (!_pp_var_multiplicity) return;
    for (arb_size_type ix = 0; ix < 2; ++ix) {
        for (arb_size_type iy = 0; iy < _pp_var_width; ++iy) {
            pp->state_vars[ix][iy] *= _pp_var_multiplicity[iy];
        }
    }
}

static void advance_state(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type celsius = _pp_var_temperature_degC[node_indexi_];
        arb_value_type ba_1_, a_0_, ll9_, ll8_, ll0_, ll7_, a_1_, ll6_, ll3_, ll1_, ba_0_, ll5_, ll4_, ll2_, qt;
        ll9_ =  0.;
        ll8_ =  0.;
        ll7_ =  0.;
        ll6_ =  0.;
        ll5_ =  0.;
        ll4_ =  0.;
        ll3_ =  0.;
        ll2_ =  0.;
        ll1_ =  0.;
        ll0_ =  0.;
        qt = exp( 1.0986122886681098*((celsius- 24.0)* 0.10000000000000001));
        ll0_ =  1.0/( 1.0+exp( -(v-_pp_var_sha- 7.5999999999999996)* 0.071428571428571425));
        ll1_ = exp(_pp_var_zetam*(v-_pp_var_vhalfm));
        ll2_ = exp(_pp_var_zetam*_pp_var_gmm*(v-_pp_var_vhalfm));
        a_0_ =  -1.0*qt*_pp_var_a0m*( 1.0+ll1_)/ll2_;
        ba_0_ = ll0_*qt*_pp_var_a0m*( 1.0+ll1_)/ll2_/a_0_;
        ll6_ = a_0_*_pp_var_dt;
        ll7_ = ( 1.0+ 0.5*ll6_)/( 1.0- 0.5*ll6_);
        _pp_var_m[i_] =  -ba_0_+(_pp_var_m[i_]+ba_0_)*ll7_;
        ll3_ =  1.0/( 1.0+exp((v-_pp_var_shi+ 47.399999999999999)* 0.16666666666666666));
        ll4_ = exp(_pp_var_zetah*(v-_pp_var_vhalfh));
        ll5_ = exp(_pp_var_zetah*_pp_var_gmh*(v-_pp_var_vhalfh));
        a_1_ =  -1.0*qt*_pp_var_a0h*( 1.0+ll4_)/ll5_;
        ba_1_ = ll3_*qt*_pp_var_a0h*( 1.0+ll4_)/ll5_/a_1_;
        ll8_ = a_1_*_pp_var_dt;
        ll9_ = ( 1.0+ 0.5*ll8_)/( 1.0- 0.5*ll8_);
        _pp_var_h[i_] =  -ba_1_+(_pp_var_h[i_]+ba_1_)*ll9_;
    }
}

static void compute_currents(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto ion_k_indexi_ = _pp_var_ion_k_index[i_];
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type conductivity_ = 0;
        arb_value_type current_ = 0;
        arb_value_type conductivity_k_ = 0;
        arb_value_type ek = _pp_var_ion_k.reversal_potential[ion_k_indexi_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type ik = 0;
        ik = _pp_var_gbar[i_]*_pp_var_m[i_]*_pp_var_h[i_]*(v-ek);
        conductivity_k_ = _pp_var_gbar[i_]*_pp_var_m[i_]*_pp_var_h[i_];
        current_ = ik;
        conductivity_ = _pp_var_gbar[i_]*_pp_var_m[i_]*_pp_var_h[i_];
        _pp_var_vec_g[node_indexi_] = fma(10.0*_pp_var_weight[i_], conductivity_, _pp_var_vec_g[node_indexi_]);
        _pp_var_vec_i[node_indexi_] = fma(10.0*_pp_var_weight[i_], current_, _pp_var_vec_i[node_indexi_]);
        _pp_var_ion_k.conductivity[ion_k_indexi_] = fma(10.0*_pp_var_weight[i_], conductivity_k_, _pp_var_ion_k.conductivity[ion_k_indexi_]);
        _pp_var_ion_k.current_density[ion_k_indexi_] = fma(10.0*_pp_var_weight[i_], ik, _pp_var_ion_k.current_density[ion_k_indexi_]);
    }
}

static void write_ions(arb_mechanism_ppack* pp) {
}

static void apply_events(arb_mechanism_ppack*, arb_deliverable_event_stream*) {}

static void post_event(arb_mechanism_ppack*) {}
#undef PPACK_IFACE_BLOCK
} // namespace kernel_kamt
} // namespace default_catalogue
} // namespace arb

extern "C" {
  arb_mechanism_interface* make_arb_default_catalogue_kamt_interface_multicore() {
    static arb_mechanism_interface result;
    result.partition_width = arb::default_catalogue::kernel_kamt::simd_width_;
    result.backend = arb_backend_kind_cpu;
    result.alignment = arb::default_catalogue::kernel_kamt::min_align_;
    result.init_mechanism = arb::default_catalogue::kernel_kamt::init;
    result.compute_currents = arb::default_catalogue::kernel_kamt::compute_currents;
    result.apply_events = arb::default_catalogue::kernel_kamt::apply_events;
    result.advance_state = arb::default_catalogue::kernel_kamt::advance_state;
    result.write_ions = arb::default_catalogue::kernel_kamt::write_ions;
    result.post_event = arb::default_catalogue::kernel_kamt::post_event;
    return &result;
  }
}

