#include <algorithm>
#include <cmath>
#include <cstddef>
#include <memory>
#include <arbor/mechanism_abi.h>
#include <arbor/math.hpp>

namespace arb {
namespace default_catalogue {
namespace kernel_nax {

using ::arb::math::exprelr;
using ::arb::math::safeinv;
using ::std::abs;
using ::std::cos;
using ::std::exp;
using ::std::max;
using ::std::min;
using ::std::pow;
using ::std::sin;
using ::std::sqrt;
using ::std::tanh;

static constexpr unsigned simd_width_ = 1;
static constexpr unsigned min_align_ = std::max(alignof(arb_value_type), alignof(arb_index_type));
using ::std::log;

#define PPACK_IFACE_BLOCK \
[[maybe_unused]] auto _pp_var_width                                                 = pp->width;\
[[maybe_unused]] auto _pp_var_n_detectors                                           = pp->n_detectors;\
[[maybe_unused]] auto _pp_var_dt                                                    = pp->dt;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_vec_ci                       = pp->vec_ci;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_v                        = pp->vec_v;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_i                        = pp->vec_i;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_g                        = pp->vec_g;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_temperature_degC             = pp->temperature_degC;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_diam_um                      = pp->diam_um;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_area_um2                     = pp->area_um2;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_time_since_spike             = pp->time_since_spike;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_node_index                   = pp->node_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_peer_index                   = pp->peer_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_multiplicity                 = pp->multiplicity;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_weight                       = pp->weight;\
[[maybe_unused]] auto& _pp_var_events                                               = pp->events;\
[[maybe_unused]] auto _pp_var_mechanism_id                                          = pp->mechanism_id;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_contiguous               = pp->index_constraints.n_contiguous;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_constant                 = pp->index_constraints.n_constant;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_independent              = pp->index_constraints.n_independent;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_none                     = pp->index_constraints.n_none;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_contiguous  = pp->index_constraints.contiguous;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_constant    = pp->index_constraints.constant;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_independent = pp->index_constraints.independent;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_none        = pp->index_constraints.none;\
[[maybe_unused]] auto _pp_var_tha = pp->globals[0];\
[[maybe_unused]] auto _pp_var_qa = pp->globals[1];\
[[maybe_unused]] auto _pp_var_Ra = pp->globals[2];\
[[maybe_unused]] auto _pp_var_Rb = pp->globals[3];\
[[maybe_unused]] auto _pp_var_thi1 = pp->globals[4];\
[[maybe_unused]] auto _pp_var_thi2 = pp->globals[5];\
[[maybe_unused]] auto _pp_var_qd = pp->globals[6];\
[[maybe_unused]] auto _pp_var_qg = pp->globals[7];\
[[maybe_unused]] auto _pp_var_mmin = pp->globals[8];\
[[maybe_unused]] auto _pp_var_hmin = pp->globals[9];\
[[maybe_unused]] auto _pp_var_q10 = pp->globals[10];\
[[maybe_unused]] auto _pp_var_Rg = pp->globals[11];\
[[maybe_unused]] auto _pp_var_Rd = pp->globals[12];\
[[maybe_unused]] auto _pp_var_thinf = pp->globals[13];\
[[maybe_unused]] auto _pp_var_qinf = pp->globals[14];\
[[maybe_unused]] auto const * const * _pp_var_random_numbers = pp->random_numbers;\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_m = pp->state_vars[0];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_h = pp->state_vars[1];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_v = pp->state_vars[2];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_sh = pp->parameters[0];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_gbar = pp->parameters[1];\
[[maybe_unused]] auto& _pp_var_ion_na = pp->ion_states[0];\
[[maybe_unused]] auto* __restrict__ _pp_var_ion_na_index = pp->ion_states[0].index;\
//End of IFACEBLOCK


// interface methods
static void init(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type u, b, a, ll0_, ll1_;
        ll1_ =  0.;
        ll0_ =  0.;
        u = _pp_var_tha+_pp_var_sh[i_]-v;
        a = _pp_var_Ra*_pp_var_qa*exprelr(u/_pp_var_qa);
        ll0_ =  -u;
        b = _pp_var_Rb*_pp_var_qa*exprelr(ll0_/_pp_var_qa);
        _pp_var_m[i_] = a/(a+b);
        u = _pp_var_thi1+_pp_var_sh[i_]-v;
        a = _pp_var_Rd*_pp_var_qd*exprelr(u/_pp_var_qd);
        ll1_ =  -u;
        b = _pp_var_Rg*_pp_var_qg*exprelr(ll1_/_pp_var_qg);
        _pp_var_h[i_] =  1.0/( 1.0+exp((v-_pp_var_thinf-_pp_var_sh[i_])/_pp_var_qinf));
    }
    if (!_pp_var_multiplicity) return;
    for (arb_size_type ix = 0; ix < 2; ++ix) {
        for (arb_size_type iy = 0; iy < _pp_var_width; ++iy) {
            pp->state_vars[ix][iy] *= _pp_var_multiplicity[iy];
        }
    }
}

static void advance_state(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type celsius = _pp_var_temperature_degC[node_indexi_];
        arb_value_type a_1_, a_0_, u, ll3_, b_1_, ll1_, ll5_, ll4_, ll2_, ll0_, iq, a, b_0_, htau, b, hinf, minf, mtau;
        ll5_ =  0.;
        ll4_ =  0.;
        ll3_ =  0.;
        ll2_ =  0.;
        ll1_ =  0.;
        ll0_ =  0.;
        iq = exp(log(_pp_var_q10)*( -0.10000000000000001*(celsius- 24.0)));
        u = _pp_var_tha+_pp_var_sh[i_]-v;
        a = _pp_var_Ra*_pp_var_qa*exprelr(u/_pp_var_qa);
        ll0_ =  -u;
        b = _pp_var_Rb*_pp_var_qa*exprelr(ll0_/_pp_var_qa);
        mtau = max(iq/(a+b), _pp_var_mmin);
        minf = a/(a+b);
        u = _pp_var_thi1+_pp_var_sh[i_]-v;
        a = _pp_var_Rd*_pp_var_qd*exprelr(u/_pp_var_qd);
        ll1_ =  -u;
        b = _pp_var_Rg*_pp_var_qg*exprelr(ll1_/_pp_var_qg);
        htau = max(iq/(a+b), _pp_var_hmin);
        hinf =  1.0/( 1.0+exp((v-_pp_var_thinf-_pp_var_sh[i_])/_pp_var_qinf));
        a_0_ = mtau;
        b_0_ = minf;
        ll2_ =  -_pp_var_dt/a_0_;
        ll3_ = ( 1.0+ 0.5*ll2_)/( 1.0- 0.5*ll2_);
        _pp_var_m[i_] = b_0_+(_pp_var_m[i_]-b_0_)*ll3_;
        a_1_ = htau;
        b_1_ = hinf;
        ll4_ =  -_pp_var_dt/a_1_;
        ll5_ = ( 1.0+ 0.5*ll4_)/( 1.0- 0.5*ll4_);
        _pp_var_h[i_] = b_1_+(_pp_var_h[i_]-b_1_)*ll5_;
    }
}

static void compute_currents(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto ion_na_indexi_ = _pp_var_ion_na_index[i_];
        auto node_indexi_ = _pp_var_node_index[i_];
        arb_value_type current_ = 0;
        arb_value_type ena = _pp_var_ion_na.reversal_potential[ion_na_indexi_];
        arb_value_type v = _pp_var_vec_v[node_indexi_];
        arb_value_type ina = 0;
        arb_value_type conductivity_ = 0;
        arb_value_type conductivity_na_ = 0;
        arb_value_type gna;
        gna = _pp_var_gbar[i_]*_pp_var_m[i_]*_pp_var_m[i_]*_pp_var_m[i_]*_pp_var_h[i_];
        ina = gna*(v-ena);
        conductivity_na_ = gna;
        current_ = ina;
        conductivity_ = gna;
        _pp_var_vec_i[node_indexi_] = fma(10.0*_pp_var_weight[i_], current_, _pp_var_vec_i[node_indexi_]);
        _pp_var_ion_na.current_density[ion_na_indexi_] = fma(10.0*_pp_var_weight[i_], ina, _pp_var_ion_na.current_density[ion_na_indexi_]);
        _pp_var_vec_g[node_indexi_] = fma(10.0*_pp_var_weight[i_], conductivity_, _pp_var_vec_g[node_indexi_]);
        _pp_var_ion_na.conductivity[ion_na_indexi_] = fma(10.0*_pp_var_weight[i_], conductivity_na_, _pp_var_ion_na.conductivity[ion_na_indexi_]);
    }
}

static void write_ions(arb_mechanism_ppack* pp) {
}

static void apply_events(arb_mechanism_ppack*, arb_deliverable_event_stream*) {}

static void post_event(arb_mechanism_ppack*) {}
#undef PPACK_IFACE_BLOCK
} // namespace kernel_nax
} // namespace default_catalogue
} // namespace arb

extern "C" {
  arb_mechanism_interface* make_arb_default_catalogue_nax_interface_multicore() {
    static arb_mechanism_interface result;
    result.partition_width = arb::default_catalogue::kernel_nax::simd_width_;
    result.backend = arb_backend_kind_cpu;
    result.alignment = arb::default_catalogue::kernel_nax::min_align_;
    result.init_mechanism = arb::default_catalogue::kernel_nax::init;
    result.compute_currents = arb::default_catalogue::kernel_nax::compute_currents;
    result.apply_events = arb::default_catalogue::kernel_nax::apply_events;
    result.advance_state = arb::default_catalogue::kernel_nax::advance_state;
    result.write_ions = arb::default_catalogue::kernel_nax::write_ions;
    result.post_event = arb::default_catalogue::kernel_nax::post_event;
    return &result;
  }
}

