#include <algorithm>
#include <cmath>
#include <cstddef>
#include <memory>
#include <arbor/mechanism_abi.h>
#include <arbor/math.hpp>

namespace arb {
namespace stochastic_catalogue {
namespace kernel_calcium_based_synapse {

using ::arb::math::exprelr;
using ::arb::math::safeinv;
using ::std::abs;
using ::std::cos;
using ::std::exp;
using ::std::max;
using ::std::min;
using ::std::pow;
using ::std::sin;
using ::std::sqrt;
using ::std::tanh;

static constexpr unsigned simd_width_ = 1;
static constexpr unsigned min_align_ = std::max(alignof(arb_value_type), alignof(arb_index_type));
using ::std::log;

#define PPACK_IFACE_BLOCK \
[[maybe_unused]] auto _pp_var_width                                                 = pp->width;\
[[maybe_unused]] auto _pp_var_n_detectors                                           = pp->n_detectors;\
[[maybe_unused]] auto _pp_var_dt                                                    = pp->dt;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_vec_ci                       = pp->vec_ci;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_v                        = pp->vec_v;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_i                        = pp->vec_i;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_vec_g                        = pp->vec_g;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_temperature_degC             = pp->temperature_degC;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_diam_um                      = pp->diam_um;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_area_um2                     = pp->area_um2;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_time_since_spike             = pp->time_since_spike;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_node_index                   = pp->node_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_peer_index                   = pp->peer_index;\
[[maybe_unused]] arb_index_type * __restrict__ _pp_var_multiplicity                 = pp->multiplicity;\
[[maybe_unused]] arb_value_type * __restrict__ _pp_var_weight                       = pp->weight;\
[[maybe_unused]] auto& _pp_var_events                                               = pp->events;\
[[maybe_unused]] auto _pp_var_mechanism_id                                          = pp->mechanism_id;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_contiguous               = pp->index_constraints.n_contiguous;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_constant                 = pp->index_constraints.n_constant;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_independent              = pp->index_constraints.n_independent;\
[[maybe_unused]] arb_size_type _pp_var_index_constraints_n_none                     = pp->index_constraints.n_none;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_contiguous  = pp->index_constraints.contiguous;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_constant    = pp->index_constraints.constant;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_independent = pp->index_constraints.independent;\
[[maybe_unused]] arb_index_type* __restrict__ _pp_var_index_constraints_none        = pp->index_constraints.none;\
[[maybe_unused]] auto _pp_var_rho_star = pp->globals[0];\
[[maybe_unused]] auto const * const * _pp_var_random_numbers = pp->random_numbers;\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_c = pp->state_vars[0];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_rho = pp->state_vars[1];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_one_over_tau = pp->state_vars[2];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_one_over_tau_Ca = pp->state_vars[3];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_sigma_over_sqrt_tau = pp->state_vars[4];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_rho_0 = pp->parameters[0];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_tau = pp->parameters[1];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_gamma_p = pp->parameters[2];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_theta_p = pp->parameters[3];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_gamma_d = pp->parameters[4];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_theta_d = pp->parameters[5];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_sigma = pp->parameters[6];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_C_pre = pp->parameters[7];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_C_post = pp->parameters[8];\
[[maybe_unused]] arb_value_type* __restrict__ _pp_var_tau_Ca = pp->parameters[9];\
//End of IFACEBLOCK


// interface methods
static void init(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        _pp_var_c[i_] =  0.;
        _pp_var_rho[i_] = _pp_var_rho_0[i_];
        _pp_var_one_over_tau[i_] =  1.0/_pp_var_tau[i_];
        _pp_var_one_over_tau_Ca[i_] =  1.0/_pp_var_tau_Ca[i_];
        _pp_var_sigma_over_sqrt_tau[i_] = _pp_var_sigma[i_]/exp(log(_pp_var_tau[i_])* 0.5);
    }
    if (!_pp_var_multiplicity) return;
    for (arb_size_type ix = 0; ix < 2; ++ix) {
        for (arb_size_type iy = 0; iy < _pp_var_width; ++iy) {
            pp->state_vars[ix][iy] *= _pp_var_multiplicity[iy];
        }
    }
}

static void advance_state(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        arb_value_type hsd, hsp, w_0_, s_0_;
        s_0_ = exp(log(_pp_var_dt)* 0.5);
        w_0_ = s_0_*_pp_var_random_numbers[0][i_];
        hsp = ((arb_value_type)((_pp_var_c[i_]-_pp_var_theta_p[i_])>=0.));
        hsd = ((arb_value_type)((_pp_var_c[i_]-_pp_var_theta_d[i_])>=0.));
        _pp_var_rho[i_] = _pp_var_rho[i_]+( -_pp_var_rho[i_]*( 1.0-_pp_var_rho[i_])*(_pp_var_rho_star-_pp_var_rho[i_])+_pp_var_gamma_p[i_]*( 1.0-_pp_var_rho[i_])*hsp-_pp_var_gamma_d[i_]*_pp_var_rho[i_]*hsd)*_pp_var_one_over_tau[i_]*_pp_var_dt+exp(log(hsp+hsd)* 0.5)*_pp_var_sigma_over_sqrt_tau[i_]*w_0_;
        _pp_var_c[i_] = _pp_var_c[i_]+ -_pp_var_c[i_]*_pp_var_one_over_tau_Ca[i_]*_pp_var_dt;
    }
}

static void compute_currents(arb_mechanism_ppack* pp) {
}

static void write_ions(arb_mechanism_ppack* pp) {
}

static void apply_events(arb_mechanism_ppack* pp, arb_deliverable_event_stream* stream_ptr) {
    PPACK_IFACE_BLOCK;
    auto [begin_, end_] = *stream_ptr;
    for (; begin_<end_; ++begin_) {
        [[maybe_unused]] auto [i_, weight] = *begin_;
        _pp_var_c[i_] = _pp_var_c[i_]+_pp_var_C_pre[i_];
    }
}

static void post_event(arb_mechanism_ppack* pp) {
    PPACK_IFACE_BLOCK;
    for (arb_size_type i_ = 0; i_ < _pp_var_width; ++i_) {
        auto node_index_i_ = _pp_var_node_index[i_];
        auto cid_          = _pp_var_vec_ci[node_index_i_];
        auto offset_       = _pp_var_n_detectors * cid_;
        for (auto c = 0; c < _pp_var_n_detectors; c++) {
            auto time = _pp_var_time_since_spike[offset_ + c];
            if (time >= 0) {
                _pp_var_c[i_] = _pp_var_c[i_]+_pp_var_C_post[i_];
            }
        }
    }
}
#undef PPACK_IFACE_BLOCK
} // namespace kernel_calcium_based_synapse
} // namespace stochastic_catalogue
} // namespace arb

extern "C" {
  arb_mechanism_interface* make_arb_stochastic_catalogue_calcium_based_synapse_interface_multicore() {
    static arb_mechanism_interface result;
    result.partition_width = arb::stochastic_catalogue::kernel_calcium_based_synapse::simd_width_;
    result.backend = arb_backend_kind_cpu;
    result.alignment = arb::stochastic_catalogue::kernel_calcium_based_synapse::min_align_;
    result.init_mechanism = arb::stochastic_catalogue::kernel_calcium_based_synapse::init;
    result.compute_currents = arb::stochastic_catalogue::kernel_calcium_based_synapse::compute_currents;
    result.apply_events = arb::stochastic_catalogue::kernel_calcium_based_synapse::apply_events;
    result.advance_state = arb::stochastic_catalogue::kernel_calcium_based_synapse::advance_state;
    result.write_ions = arb::stochastic_catalogue::kernel_calcium_based_synapse::write_ions;
    result.post_event = arb::stochastic_catalogue::kernel_calcium_based_synapse::post_event;
    return &result;
  }
}

